# OUTDATED AND NOT USED ANYMORE
# Pull from https://github.com/poljar/matrix-nio/pull/451 temporarily.
from nio.responses import DirectRoomsErrorResponse, DirectRoomsResponse

__all__ = (
    "DirectRoomsErrorResponse",
    "DirectRoomsResponse",
)
